//
//  IFS3DView.h
//  IFS3D
//
//  Copyright (c) 2001 David Leppik.
//

// This is the main view class, and the one which the screen saver program looks for.

#import "IFS3DView.h"
#import "PointShape.h"
#import "Raindrop.h"
#import "ColorTriangle.h"

@implementation IFS3DView

int frame=0;

static float eyeSeparationMultiplier = 0.005;

//#########################################################################
//  Generic screen saver stuff
//#########################################################################

- (BOOL)hasConfigureSheet {
    return YES;
}

- (NSWindow*)configureSheet {
    return [[IFS3DPreferences sharedInstance] window];
}

- (NSTimeInterval)animationTimeInterval { return MAX_FRAMERATE_SECS; }

//----------------------------------------------------------------------------------
//  initialization
//----------------------------------------------------------------------------------
- (id)initWithFrame:(NSRect)frameRect isPreview:(BOOL)isPreview
{    
    self = [super initWithFrame:frameRect isPreview:isPreview];
    if (self) {
        // Cocoa OpenGL classes aren't documented yet and am
        // not familiar with the various hardware configurations,
        // so these settings represent a guess of what I can get away with.
        NSOpenGLPixelFormatAttribute bestAttribs[] =
        {
            NSOpenGLPFAAccelerated,
            NSOpenGLPFADepthSize, 16,
            NSOpenGLPFAMinimumPolicy,
            NSOpenGLPFAClosestPolicy,
            0
        };
	NSOpenGLPixelFormatAttribute mediocreAttribs[] =
        {
            NSOpenGLPFADepthSize, 16,
            NSOpenGLPFAMinimumPolicy,
            NSOpenGLPFAClosestPolicy,
            0
        };
	NSOpenGLPixelFormatAttribute fallbackAttribs[] =
        {
            0
        };

        
	NSOpenGLPixelFormat *format = [[NSOpenGLPixelFormat alloc] initWithAttributes:bestAttribs];
        if (!format) {
            format = [[NSOpenGLPixelFormat alloc] initWithAttributes:mediocreAttribs];
            if (!format) {
                format = [[NSOpenGLPixelFormat alloc] initWithAttributes:fallbackAttribs];
            }
        }
        
        theGLView = [[[NSOpenGLView alloc] initWithFrame:NSZeroRect pixelFormat:format] autorelease];
        [self addSubview:theGLView];
        [format release];
        
        viewRotation=0;
        estimatedFrameRate = MAX_FRAMERATE_SECS;
        lastFrameTime = nil;
    }
    return self;
}

- (void)setFrameSize:(NSSize)newSize
{
    screenSize = newSize;
    [super setFrameSize:newSize];
    [theGLView setFrameSize:newSize];
    isGLViewInited = NO;
}

- (void)initGLView
{
    glViewport( 0,0, screenSize.width, screenSize.height );
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
    gluPerspective( 90.0, (double)(screenSize.width / screenSize.height), 0.1, 10000.0);
    glMatrixMode( GL_MODELVIEW );
    glEnable(GL_FOG);
    glFogf(GL_FOG_DENSITY, 0.5);
    glFogf(GL_FOG_END, 4);
    
    glClearColor( 0.0, 0.0, 0.0, 0.0 );
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
}


//--------------------------------------------------------------------------------------
//  Animate
//--------------------------------------------------------------------------------------

- (void)drawScope {
    float i;

    glLineWidth( 5 );
    glBegin( GL_LINES );
        glVertex3f( 1.0, 0.0, 1.0 );
        glVertex3f(-1.0, 0.0, 1.0 );
        
        glVertex3f( 0.0, 1.0, 1.0 );
        glVertex3f( 0.0,-1.0, 1.0 );
    glEnd();
    
    glColor4f( 1, 1, 0, 1.0 );
    glBegin( GL_POINTS );
        for (i=0; i<estimatedFrameRate; i+=0.01) {
            glVertex3f( i*5, 0, 0 );
        }
    glEnd();

    
    /*
    glBegin( GL_LINE_LOOP );
        glVertex3f(-1.0,-1.0,-1.0 );
        glVertex3f( 1.0,-1.0,-1.0 );
        glVertex3f( 1.0, 1.0,-1.0 );
        glVertex3f(-1.0, 1.0,-1.0 );
    glEnd();

    glBegin( GL_LINE_LOOP );
        glVertex3f( 1.0,-1.0,-1.0 );
        glVertex3f( 1.0,-1.0, 1.0 );
        glVertex3f( 1.0, 1.0, 1.0 );
        glVertex3f( 1.0, 1.0,-1.0 );
    glEnd();
    */
}

- (void)animateOneFrame
{
  frame++;
  
  [[theGLView openGLContext] makeCurrentContext];
  if (!isGLViewInited) {
    [self initGLView];
    isGLViewInited = YES;
  }
  [self doAnimation];
  [[[IFS3DDefaults sharedInstance] fractal] updateBranches];
  [self updateFrameRateEstimate];
}

- (void)updateFrameRateEstimate {
  NSDate *now = [[NSDate date] retain];
  if (lastFrameTime) {
    estimatedFrameRate = -[lastFrameTime timeIntervalSinceNow];
    [lastFrameTime release];
  }
  else {
    estimatedFrameRate = MAX_FRAMERATE_SECS;
  }
  lastFrameTime = now;
}

- (void)drawFractal
{
  IFS3D *fractal = [[IFS3DDefaults sharedInstance] fractal];
  int reflections = [[IFS3DDefaults sharedInstance] intForKey:@"view.reflections"];
  
  // We don't really do reflections, it's more interesting to rotate
  // in reflection-like ways.
  
    // Each option builds on the last, so no break statements.
    switch(reflections) {
      case 2: // Something similar to reflecting on the Y=X plane
        glPushMatrix();
          glRotatef(90, 0,0,1);
          glRotatef(90, 1,0,0); 
            [fractal draw3DWithView:self];
            glPushMatrix();
              glRotatef(180, 0,0,1);
              glRotatef(180, 1,0,0); 
              [fractal draw3DWithView:self];
            glPopMatrix();
        glPopMatrix();
      case 1: // Rotate 90-degrees, similar to rotating on Y-axis
        glPushMatrix();
          glRotatef(180, 0,1,0);
            [fractal draw3DWithView:self];
        glPopMatrix();
      default:
        [fractal draw3DWithView:self];
    }
}

- (void)doAnimation
{
  IFS3DDefaults *defaults = [IFS3DDefaults sharedInstance];
  IFS3D *fractal = [defaults fractal];
  float rotationDefault = [defaults floatForKey:@"view.rotationSpeed"];
  
  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
  
  viewRotation += (rotationDefault/1000.0) * estimatedFrameRate / MAX_FRAMERATE_SECS;
  
  [[fractal shape] callGlParams];
    
  if ( [[fractal shape]  isStereo] ) {
    float separation = [defaults floatForKey:@"view.eyeSeparation"] *eyeSeparationMultiplier;
    // Draw right eye view
    [[fractal shape]  setEye:RIGHT];
    glLoadIdentity();
    gluLookAt(-separation, 1, -3,  0.0, 0.0,10.0,   0.0, 1.0, 0.0 );
    glRotatef( viewRotation, 0.0, 1.0, 0.0 );
    glPushMatrix();
    [self drawFractal];
    glPopMatrix();
    
    // Draw left eye view
    [[fractal shape]  setEye:LEFT];
    glLoadIdentity();
    gluLookAt( separation, 1, -3,  0.0, 0.0,10.0,   0.0, 1.0, 0.0 );
    glRotatef( viewRotation, 0.0, 1.0, 0.0 );
    glPushMatrix();
    [self drawFractal];
    glPopMatrix();
  }
  else {
    // Draw standard view
    glLoadIdentity();
    // Camera at (0,1,-3), look at (0,0,0), up is +y
    gluLookAt( 0, 1, -3,   0.0, 0.0, 0.0,   0.0, 1.0, 0.0 );
    glRotatef( viewRotation, 0.0, 1.0, 0.0 );
    [self drawFractal];
  }
  
  glFlush();
}

//------------------------------------------------------------------------------------
//  Clean-up
//------------------------------------------------------------------------------------
- (void)dealloc
{
    [theGLView removeFromSuperview];
    if (lastFrameTime) {
        [lastFrameTime release];
    }
    [super dealloc];
}

@end
